"""Defines the data types"""

from enum import Enum
from typing import NamedTuple, Dict, List


class Objective(Enum):
    """The routing objective"""

    MINIMIZE_TIME = "minimize travel time"
    MINIMIZE_DISTANCE = "minimize travel distance"
    MAXIMIZE_DELIVERIES = "maximize number of orders delivered"


class Vehicle(NamedTuple):
    """The data structure that contains vehicle information"""

    name: str  # The name of the vehicle
    shift_start: int  # The departure time of the vehicle (in minutes), 0 if non-applicable
    shift_end: int  # The return deadline of the vehicle (in minutes), 0 if non-applicable
    break_start: int  # The start time of the break (in minutes), 0 if non-applicable
    break_end: int  # The end time of the break (in minutes), 0 if non-applicable
    capacity: int  # The capacity of the vehicle, 0 if non-applicable


class Route:
    """Contains information of a route

    Attributes:
        route: The list of order IDs of locations visited
        departure: Time of departure from the depot
        return_time: Time of returning to the depot
        arrival_times: The arrival times at each order
        travel_time: The total travel time excluding wait times and service times
        travel_distance: The total travel distance
        trip_duration: The total duration of the trip

    """

    def __init__(
            self,
            route: List[int],
            departure: int,
            return_time: int,
            arrival_times: List[int],
            travel_time: int,
            travel_distance: int,
            **_
    ):
        self.route = route
        self.departure = departure
        self.return_time = return_time
        self.arrival_times = arrival_times
        self.travel_time = travel_time
        self.travel_distance = travel_distance
        self.trip_duration = self.return_time - self.departure

    def __str__(self):
        # Determine width of the first column
        order_id_len = max(len(str(order)) for order in self.route)
        first_column_len = max(8, order_id_len)
        res = "order_id".rjust(first_column_len, " ")

        # Finish header row
        res += "   arrival_time\n"
        res += "\n".rjust(len(res), "-")

        # Add the rest of the rows
        for order, time in zip(self.route, self.arrival_times):
            hour, minute = time // 60, time % 60
            res += f'{order:>{first_column_len}d}       {hour:02}:{minute:02}\n'

        # Add statistics
        res += f"\nDeparture Time: {self.departure} min"
        res += f"\nReturn Time: {self.return_time} min"
        res += f"\nNumber of orders served: {len(self.route)}"
        res += f"\nTotal Travel Time: {self.travel_time} min"
        res += f"\nTotal Distance Travelled: {self.travel_distance / 1000:.2f} km"
        hour, minute = self.trip_duration // 60, self.trip_duration % 60
        res += f"\nTrip Duration: {hour:02}h{minute:02}min"
        return res

    def to_dict(self):
        return vars(self)

    @classmethod
    def from_dict(cls, dict_obj: Dict):
        return cls(**dict_obj)



"""Defines the client that composes requests to the ForeQast API."""

import warnings
from typing import Dict
from urllib.parse import urlencode

import requests
from pandas import DataFrame

from foreqast.vrp import Vehicle, VRP

_LOCAL_BASE_URL = "http://127.0.0.1:5000/api"
_SERVER_BASE_URL = "http://54.226.79.68:5000/api"
_DEFAULT_TIMEOUT = 120


class Client:
    """The request client for the ForeQast API"""

    def __init__(
            self,
            api_key: str,
            base_url: str = _SERVER_BASE_URL,
            timeout: int = _DEFAULT_TIMEOUT
    ):
        self._api_key = api_key
        self._base_url = base_url
        self._session = requests.Session()

        self._timeout = timeout
        self._requests_kwargs = {
            "headers": {
                "Content-type": "application/json",
                "Authorization": self._api_key,
            },
            "timeout": self._timeout,
        }

    def request(
            self,
            method: str,
            url: str,
            params: Dict = None,
            request_kwargs: Dict = None
    ):
        """Compose and send a request"""

        request_method = self._session.post if method == "POST" else self._session.get

        request_kwargs = request_kwargs or {}
        final_requests_kwargs = dict(self._requests_kwargs, **request_kwargs)

        if params and method == "POST":
            final_requests_kwargs["json"] = params

        if params and method == "GET":
            url = url + "?" + urlencode(list(params.items()))

        try:
            response = request_method(self._base_url + url, **final_requests_kwargs)
            if response.status_code != 200:
                raise ValueError(response.text)
            body = response.json()
            if body.get("status") == "ERROR":
                raise ValueError(body.get("message"))
            elif body.get("status") == "WARNING":
                warnings.warn(body.get("message"))
            return body["result"]
        except requests.exceptions.Timeout:
            return {}

    def vrp(
            self,
            depot_address: str,
            orders: DataFrame,
            vehicles: Dict[str, Vehicle],
    ) -> VRP:
        """Constructs a VRP instance"""
        return VRP(self, depot_address, orders, vehicles)



from setuptools import setup, find_packages

setup(
    name="foreqast-client",
    packages=find_packages(),
    version="1.0.0",
    description="Python client for requests to the ForeQast API.",
    author="Qiyuan Cai",
    author_email="astral@foreqast.ai",
    install_requires=[
        "pandas",
        "requests"
    ],
)



"""Functions related to the VRP."""

from typing import Dict, List

from pandas import DataFrame

from foreqast.vrp.data_types import Vehicle, Objective, Route
from foreqast.vrp.utils import get_coordinates, compute_adj_matrices

DEPOT_INDEX = 0


class VRP:
    """Container for problem definition and constraints

    Attributes:
        depot_address: The address of the depot
        depot_coordinates: The coordinates of the depot
        orders: The list of orders to plan routes for
        vehicles: The list of vehicles to plan routes for
        time_matrix: The travel time matrix
        distance_matrix: The travel distance matrix

    """

    def __init__(
            self,
            client,
            depot_address: str,
            orders: DataFrame,
            vehicles: Dict[str, Vehicle],
    ):
        self.client = client
        self.depot_address = depot_address
        self.depot_coordinates = get_coordinates(self.client, depot_address)
        self.orders = orders
        self.vehicles = vehicles

        indices = [DEPOT_INDEX] + orders.index.to_list()
        coordinates = [self.depot_coordinates] + orders['coordinates'].to_list()
        self.time_matrix, self.distance_matrix = compute_adj_matrices(
            self.client, indices, coordinates)

    def cluster(
            self,
            objective: Objective,
            vehicles: List[str] = None,
            quantum=True
    ):
        """Attach cluster assignments to the orders

        Args:
            objective: The objective of the routing
            vehicles: The vehicles used in clustering, use all vehicles by default
            quantum: Whether to use the quantum algorithm for the clustering

        """
        selected_vehicles = {
            k: v for k, v in self.vehicles.items() if k in vehicles
        } if vehicles is not None else self.vehicles
        bound_capacity = objective in (Objective.MINIMIZE_DISTANCE, Objective.MINIMIZE_TIME)
        body = self.client.request(
            "POST", "/vrp/cluster", {
                "orders": self.orders.to_json(orient='split'),
                "vehicles": selected_vehicles,
                "time_matrix": self.time_matrix.to_json(orient='split'),
                "bound_capacity": bound_capacity,
                "quantum": quantum,
            }
        )
        self.orders = self.orders.assign(cluster=body['clusters'])

    def plan_routes(self, objective: Objective, vehicles: List[str] = None):
        """Plans the routes for each vehicle

        Args:
            objective: The objective of the routing
            vehicles: The list of vehicles to plan routes for, use all vehicles if not specified

        Returns:
            A dictionary containing the Route object generated for each vehicle

        """

        available_vehicles = {k: v for k, v in self.vehicles.items() if k in self.orders['cluster'].unique()}
        selected_vehicles = {
            k: v for k, v in available_vehicles.items() if k in vehicles
        } if vehicles is not None else available_vehicles
        body = self.client.request(
            "POST", "/vrp/plan-routes", {
                "orders": self.orders.to_json(orient='split'),
                "vehicles": selected_vehicles,
                "time_matrix": self.time_matrix.to_json(orient='split'),
                "distance_matrix": self.distance_matrix.to_json(orient='split'),
                "objective": objective.name
            }
        )
        return dict((k, Route.from_dict(v)) for k, v in body['routes'].items())

    def set_priority(self, order_id: int, priority: int = 1):
        """Sets the priority of an order to

        Args:
            order_id: The id of the order
            priority: The priority of the order, must be 1 or 0

        """

        if 'priority' not in self.orders:
            self.orders = self.orders.assign(priority=0)
        self.orders.loc[order_id, 'priority'] = int(priority)

    def solve_vrp(
            self,
            objective: Objective,
            vehicles: List[str] = None,
            balancing_alpha=0,
            use_quantum=True
    ):
        """Plans the routes for each vehicle

        Args:
            objective: The objective of the routing
            vehicles: The list of vehicles to plan routes for, use all vehicles if not specified
            balancing_alpha: The alpha value for balancing the clusters, typically between 0 and 8.
            use_quantum: Whether to use the quantum algorithm for the clustering

        Returns:
            A dictionary containing the Route object generated for each vehicle

        """

        selected_vehicles = {
            k: v for k, v in self.vehicles.items() if k in vehicles
        } if vehicles is not None else self.vehicles

        body = self.client.request(
            "POST", "/vrp/solve-vrp", {
                "orders": self.orders.to_json(orient='split'),
                "vehicles": selected_vehicles,
                "time_matrix": self.time_matrix.to_json(orient='split'),
                "distance_matrix": self.distance_matrix.to_json(orient='split'),
                "balancing_alpha": balancing_alpha,
                "bound_capacity": objective != Objective.MAXIMIZE_DELIVERIES,
                "objective": objective.name,
                "use_quantum": use_quantum
            }
        )
        self.orders = self.orders.assign(cluster=body['clusters'])
        return dict((k, Route.from_dict(v)) for k, v in body['routes'].items())



"""Utility functions for the VRP module"""
from io import StringIO
from typing import Dict, List

import pandas as pd
from pandas import DataFrame

from foreqast.vrp.data_types import Vehicle


def load_orders(file_name: str) -> DataFrame:
    """Reads the file containing orders"""

    df = pd.read_excel(
        file_name,
        index_col=0,
        converters={
            'coordinates': eval
        }
    )

    orders = df[['coordinates']].copy()

    if 'date' in df.columns:
        orders['date'] = df['date']

    if 'tw_start' in df.columns:
        orders['tw_start'] = df['tw_start'].map(str).map(__convert_time_to_minutes)
        orders['tw_end'] = df['tw_end'].map(str).map(__convert_time_to_minutes)

    if 'service_time' in df.columns:
        orders['service_time'] = df['service_time'].map(lambda x: 0 if pd.isna(x) else int(x))

    if 'load' in df.columns:
        orders['load'] = df['load'].map(lambda x: 0 if pd.isna(x) else int(x))

    if 'priority' in df.columns:
        orders['priority'] = df['priority'].map(int)

    return orders


def load_vehicles(file_name: str) -> Dict[str, Vehicle]:
    """Reads the file containing vehicle information"""

    df = pd.read_excel(file_name)

    if 'Shift Start' not in df.columns:
        df['Shift Start'] = 0
        df['Shift End'] = 0
    else:
        df['Shift Start'] = df['Shift Start'].map(str).map(__convert_time_to_minutes)
        df['Shift End'] = df['Shift End'].map(str).map(__convert_time_to_minutes)

    if 'Break Start' not in df.columns:
        df['Break Start'] = 0
        df['Break End'] = 0
    else:
        df['Break Start'] = df['Break Start'].map(str).map(__convert_time_to_minutes)
        df['Break End'] = df['Break End'].map(str).map(__convert_time_to_minutes)

    if 'Capacity' not in df.columns:
        df['Capacity'] = 0
    else:
        df['Capacity'] = df['Capacity'].map(int)

    vehicles = {}
    for _, row in df.iterrows():
        vehicles[row['Vehicle']] = Vehicle(
            row['Vehicle'],
            row['Shift Start'],
            row['Shift End'],
            row['Break Start'],
            row['Break End'],
            row['Capacity']
        )

    return vehicles


def __convert_time_to_minutes(time: str) -> int:
    return 0 if pd.isna(time) else int(time[:-2]) * 60 + int(time[-2:])


def get_coordinates(client, address: str):
    """Gets the coordinates of an address in lng, lat format"""
    result = client.request(
        "GET", f"/map-services/get-coordinates", {"address": address})
    return result['coordinates']


def compute_adj_matrices(client, indices: List, coordinates: List):
    """Gets the travel time matrix and travel distance matrix"""
    body = client.request(
        "POST", f"/map-services/adj-matrices", {
            "indices": indices,
            "coordinates": coordinates
        })
    time_matrix = pd.read_json(StringIO(body['time_matrix']), orient='split')
    distance_matrix = pd.read_json(StringIO(body['distance_matrix']), orient='split')
    return time_matrix, distance_matrix
